from __future__ import annotations

from typing import TYPE_CHECKING

from .api import PulseEcoAPI
from .constants import PULSE_ECO_BASE_URL
from .models import DataValue, Overall, Sensor

if TYPE_CHECKING:
    import datetime

    import requests

    from .api.base import PulseEcoAPIBase
    from .enums import AveragePeriod, DataValueType


class PulseEcoClient:
    """High level pulse.eco client."""

    def __init__(
        self,
        city_name: str,
        auth: tuple[str, str] | None = None,
        base_url: str = PULSE_ECO_BASE_URL,
        session: requests.Session | None = None,
        pulse_eco_api: PulseEcoAPIBase | None = None,
    ) -> None:
        """Initialize the pulse.eco client.

        :param city_name: the city name
        :param auth: a tuple of (email, password), defaults to None
        :param base_url: the base URL of the API, defaults to
            'https://{city_name}.pulse.eco/rest/{end_point}'
        :param session: a requests session
            use this to customize the session and add retries, defaults to None,
        :param pulse_eco_api: a pulse.eco API wrapper, defaults to None,
            if set, the other parameters are ignored
        """
        self._pulse_eco_api: PulseEcoAPIBase
        if pulse_eco_api is None:
            self._pulse_eco_api = PulseEcoAPI(
                city_name=city_name, auth=auth, base_url=base_url, session=session
            )
        else:
            self._pulse_eco_api = pulse_eco_api

    def sensors(self) -> list[Sensor]:
        """Get all sensors for a city.

        :return: a list of sensors
        """
        return [
            Sensor.model_validate(sensor) for sensor in self._pulse_eco_api.sensors()
        ]

    def sensor(self, sensor_id: str) -> Sensor:
        """Get a sensor by it's ID.

        :param sensor_id: the unique ID of the sensor
        :return: a sensor
        """
        return Sensor.model_validate(self._pulse_eco_api.sensor(sensor_id=sensor_id))

    def data_raw(
        self,
        from_: str | datetime.datetime,
        to: str | datetime.datetime,
        type: DataValueType | None = None,
        sensor_id: str | None = None,
    ) -> list[DataValue]:
        """Get raw data for a city.

        :param from_: the start datetime of the data
            as a datetime object or an isoformat string
        :param to: the end datetime of the data
            as a datetime object or an isoformat string
        :param type: the data value type, defaults to None
        :param sensor_id: the unique ID of the sensor, defaults to None
        :return: a list of data values
        """
        return [
            DataValue.model_validate(data_value)
            for data_value in self._pulse_eco_api.data_raw(
                from_=from_,
                to=to,
                type=type,
                sensor_id=sensor_id,
            )
        ]

    def avg_data(
        self,
        period: AveragePeriod,
        from_: str | datetime.datetime,
        to: str | datetime.datetime,
        type: DataValueType,
        sensor_id: str | None = None,
    ) -> list[DataValue]:
        """Get average data for a city.

        :param period: the period of the average data
        :param from_: the start datetime of the data
            as a datetime object or an isoformat string
        :param to: the end datetime of the data
            as a datetime object or an isoformat string
        :param type: the data value type
        :param sensor_id: the unique ID of the sensor, defaults to None
        :return: a list of average data values
        """
        return [
            DataValue.model_validate(data_value)
            for data_value in self._pulse_eco_api.avg_data(
                period=period,
                from_=from_,
                to=to,
                type=type,
                sensor_id=sensor_id,
            )
        ]

    def data24h(self) -> list[DataValue]:
        """Get 24h data for a city.

        The data values are sorted ascending by their timestamp.

        :return: a list of data values for the past 24 hours
        """
        return [
            DataValue.model_validate(data_value)
            for data_value in self._pulse_eco_api.data24h()
        ]

    def current(self) -> list[DataValue]:
        """Get the last received valid data for each sensor in a city.

        Will not return sensor data older than 2 hours.

        :return: a list of current data values
        """
        return [
            DataValue.model_validate(data_value)
            for data_value in self._pulse_eco_api.current()
        ]

    def overall(self) -> Overall:
        """Get the current average data for all sensors per value for a city.

        :return: the overall data for the city
        """
        return Overall.model_validate(self._pulse_eco_api.overall())



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import datetime

    from pulseeco.api.data_types import DataValueAvg, DataValueRaw, Overall, Sensor

from abc import ABC, abstractmethod


class PulseEcoAPIBase(ABC):  # pragma: no cover
    """Low level unsafe pulse.eco API wrapper base class"""

    @abstractmethod
    def sensors(self) -> list[Sensor]: ...

    @abstractmethod
    def sensor(self, sensor_id: str) -> Sensor: ...

    @abstractmethod
    def data_raw(
        self,
        from_: str | datetime.datetime,
        to: str | datetime.datetime,
        type: str | None = None,
        sensor_id: str | None = None,
    ) -> list[DataValueRaw]: ...

    @abstractmethod
    def avg_data(
        self,
        period: str,
        from_: str | datetime.datetime,
        to: str | datetime.datetime,
        type: str,
        sensor_id: str | None = None,
    ) -> list[DataValueAvg]: ...

    @abstractmethod
    def data24h(self) -> list[DataValueRaw]: ...

    @abstractmethod
    def current(self) -> list[DataValueRaw]: ...

    @abstractmethod
    def overall(self) -> Overall: ...



from __future__ import annotations

import os
import warnings
from typing import TYPE_CHECKING, Any, cast

import requests

from pulseeco.constants import AVG_DATA_MAX_SPAN, DATA_RAW_MAX_SPAN, PULSE_ECO_BASE_URL
from pulseeco.utils import convert_datetime_to_str, split_datetime_span

from .base import PulseEcoAPIBase
from .data_types import DataValueAvg, DataValueRaw, Overall, Sensor

if TYPE_CHECKING:
    import datetime


class PulseEcoAPI(PulseEcoAPIBase):
    """Low level unsafe pulse.eco API wrapper."""

    def __init__(
        self,
        city_name: str,
        auth: tuple[str, str] | None = None,
        base_url: str = PULSE_ECO_BASE_URL,
        session: requests.Session | None = None,
    ) -> None:
        """Initialize the pulse.eco API wrapper.

        :param city_name: the city name
        :param auth: a tuple of (email, password), defaults to None
        :param base_url: the base URL of the API, defaults to
            'https://{city_name}.pulse.eco/rest/{end_point}'
        :param session: a requests session
            , use this to customize the session and add retries, defaults to None
        """
        self.city_name = city_name

        if base_url is None and "PULSE_ECO_BASE_URL" in os.environ:
            base_url = os.environ["PULSE_ECO_BASE_URL"]

        if session is not None:
            self._session = session
        else:
            self._session = requests.Session()

        if (
            auth is None
            and "PULSE_ECO_USERNAME" in os.environ
            and "PULSE_ECO_PASSWORD" in os.environ
        ):
            auth = (
                os.environ["PULSE_ECO_USERNAME"],
                os.environ["PULSE_ECO_PASSWORD"],
            )

        if auth is not None:
            self._session.auth = auth

        self._base_url = base_url

    def _base_request(
        self, end_point: str, params: dict[str, Any] | None = None
    ) -> Any:  # noqa: ANN401
        """Make a request to the PulseEco API.

        :param end_point: an end point of the API
        :param params: get parameters, defaults to None
        :return: the response json
        """
        if params is None:
            params = {}
        url = self._base_url.format(city_name=self.city_name, end_point=end_point)
        response = self._session.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def sensors(self) -> list[Sensor]:
        """Get all sensors for a city.

        :return: a list of sensors
        """
        return cast("list[Sensor]", self._base_request("sensor"))

    def sensor(self, sensor_id: str) -> Sensor:
        """Get a sensor by it's ID

        :param sensor_id: the unique ID of the sensor
        :return: a sensor
        """
        return cast(Sensor, self._base_request(f"sensor/{sensor_id}"))

    def data_raw(
        self,
        from_: str | datetime.datetime,
        to: str | datetime.datetime,
        type: str | None = None,
        sensor_id: str | None = None,
    ) -> list[DataValueRaw]:
        """Get raw data for a city.

        :param from_: the start datetime of the data
            as a datetime object or an isoformat string
        :param to: the end datetime of the data
            as a datetime object or an isoformat string
        :param type: the data value type, defaults to None
        :param sensor_id: the unique ID of the sensor, defaults to None
        :return: a list of data values
        """
        if sensor_id is None and type is None:
            warnings.warn(
                "Warning! If you encounter an error, "
                "you should probably specify either sensor_id or type.",
                stacklevel=2,
            )
        data: list[DataValueRaw] = []
        datetime_spans = split_datetime_span(from_, to, DATA_RAW_MAX_SPAN)
        for from_temp, to_temp in datetime_spans:
            params = {
                "sensorId": sensor_id,
                "type": type,
                "from": convert_datetime_to_str(from_temp),
                "to": convert_datetime_to_str(to_temp),
            }
            params = {k: v for k, v in params.items() if v is not None}
            data_value = cast(
                "list[DataValueRaw]",
                self._base_request("dataRaw", params=params),
            )
            data += data_value
        return data

    def avg_data(
        self,
        period: str,
        from_: str | datetime.datetime,
        to: str | datetime.datetime,
        type: str,
        sensor_id: str | None = None,
    ) -> list[DataValueAvg]:
        """Get average data for a city.

        :param period: the period of the average data (day, week, month)
        :param from_: the start datetime of the data
            as a datetime object or an isoformat string
        :param to: the end datetime of the data
            as a datetime object or an isoformat string
        :param type: the data value type
        :param sensor_id: the unique ID of the sensor, defaults to None
        :return: a list of average data values
        """
        if period not in {"day", "week", "month"}:
            warnings.warn(
                "Warning! Invalid value for period. "
                "Should be one of: day, week, month",
                stacklevel=2,
            )
        data: list[DataValueAvg] = []
        datetime_spans = split_datetime_span(from_, to, AVG_DATA_MAX_SPAN)
        for from_temp, to_temp in datetime_spans:
            params = {
                "sensorId": sensor_id,
                "type": type,
                "from": convert_datetime_to_str(from_temp),
                "to": convert_datetime_to_str(to_temp),
            }
            params = {k: v for k, v in params.items() if v is not None}
            data_value = cast(
                "list[DataValueAvg]",
                self._base_request(f"avgData/{period}", params=params),
            )
            data += data_value
        return data

    def data24h(self) -> list[DataValueRaw]:
        """Get 24h data for a city.

        The data values are sorted ascending by their timestamp.

        :return: a list of data values for the past 24 hours
        """
        return cast("list[DataValueRaw]", self._base_request("data24h"))

    def current(self) -> list[DataValueRaw]:
        """Get the last received valid data for each sensor in a city

        Will not return sensor data older than 2 hours.

        :return: a list of current data values
        """
        return cast("list[DataValueRaw]", self._base_request("current"))

    def overall(
        self,
    ) -> Overall:
        """Get the current average data for all sensors per value for a city.

        ## Example:

        ```python
        {
            'cityName': 'skopje',
            'values': {
                'no2': '22',
                'o3': '4',
                'pm25': '53',
                'pm10': '73',
                'temperature': '7',
                'humidity': '71',
                'pressure': '992',
                'noise_dba': '43'
            }
        }
        ```

        :return: the overall data for the city
        """
        return cast(Overall, self._base_request("overall"))



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import Iterator


def convert_datetime_to_str(datetime: datetime) -> str:
    """Convert a datetime object to a string

    :param datetime: a datetime object
    :return: an isoformat string
    """
    if datetime.tzinfo is None:
        datetime = datetime.replace(tzinfo=timezone.utc)
    return datetime.isoformat()


def split_datetime_span(
    fr: str | datetime, to: str | datetime, td: timedelta
) -> Iterator[tuple[datetime, datetime]]:
    """Split a datetime span into a list of (fr, to) datetime pairs
        with a given maximum timedelta

    :param fr: the start datetime of the span
    :param to: the end datetime of the span
    :param td: the timedelta between the datetimes
    :return: a list of datetimes
    """
    if isinstance(fr, str):
        fr = datetime.fromisoformat(fr)
    if isinstance(to, str):
        to = datetime.fromisoformat(to)
    current = fr
    prev = current
    while current + td < to:
        current += td
        yield prev, current
        prev = current + timedelta(seconds=1)
    yield prev, to



import sys
from enum import Enum

if sys.version_info < (3, 11):

    class StrEnum(str, Enum):
        def __repr__(self) -> str:
            return str.__repr__(self.value)

else:
    from enum import StrEnum


class SensorType(StrEnum):
    # MOEPP measurement station
    TYPE_0 = "0"
    # SkopjePulse LoRaWAN based sensor, version 1
    TYPE_1 = "1"
    # SkopjePulse WiFi based sensor, version 1
    TYPE_2 = "2"
    # pulse.eco WiFi based sensor, version 2
    TYPE_3 = "3"
    # pulse.eco LoRaWAN based sensor. version 2
    TYPE_4 = "4"
    # pengy device, version 1
    TYPE_20001 = "20001"
    # URAD Monitor device
    TYPE_20002 = "20002"
    # AirThings platform device
    TYPE_20003 = "20003"
    # sensor.community crowdsourced device
    TYPE_20004 = "20004"


class SensorStatus(StrEnum):
    # A user requested this location with a device ID, but not sending data yet
    REQUESTED = "REQUESTED"
    # The sensor is up and running properly
    ACTIVE = "ACTIVE"
    # The sensor is up and running properly but not yet confirmed by the community lead
    ACTIVE_UNCONFIRMED = "ACTIVE_UNCONFIRMED"
    # The sensor is registered but turned off and ignored
    INACTIVE = "INACTIVE"
    # The sensor is registered, but so far not bound to an owner
    NOT_CLAIMED = "NOT_CLAIMED"
    # The sensor is registered, but so far not bound to an owner nor confirmed by the community lead
    NOT_CLAIMED_UNCONFIRMED = "NOT_CLAIMED_UNCONFIRMED"
    # The sensor is manually removed from evidence in order to keep data sanity
    BANNED = "BANNED"


class DataValueType(StrEnum):
    NO2 = "no2"
    O3 = "o3"
    PM25 = "pm25"
    PM10 = "pm10"
    TEMPERATURE = "temperature"
    HUMIDITY = "humidity"
    PRESSURE = "pressure"
    NOISE_DBA = "noise_dba"


class AveragePeriod(StrEnum):
    DAY = "day"
    WEEK = "week"
    MONTH = "month"



from __future__ import annotations

from typing import TypedDict


class Sensor(TypedDict):
    sensorId: str
    position: str
    comments: str
    type: str
    description: str
    status: str


class DataValueBase(TypedDict):
    sensorId: str
    stamp: str
    type: str
    value: str


class DataValueRaw(DataValueBase):
    position: str


class DataValueAvg(DataValueBase):
    position: None


class OverallValues(TypedDict):
    no2: str
    o3: str
    pm25: str
    pm10: str
    temperature: str
    humidity: str
    pressure: str
    noise_dba: str


class Overall(TypedDict):
    cityName: str
    values: OverallValues



from __future__ import annotations

import datetime  # noqa: TCH003
from typing import Optional

from pydantic import BaseModel, Field

from .enums import DataValueType, SensorStatus, SensorType  # noqa: TCH001


class Sensor(BaseModel):
    sensor_id: str = Field(alias="sensorId", description="The unique ID of the sensor")
    position: str = Field(
        description="Latitude and longitude GPS coordinates of the sensor"
    )
    comments: str = Field(description="Any other comments about the sensor")
    type: SensorType = Field(description="The type of the sensor")
    description: str = Field(description="Short description / nome")
    status: SensorStatus = Field(description="The current status of the sensor")


class DataValue(BaseModel):
    sensor_id: str = Field(alias="sensorId", description="The unique ID of the sensor")
    stamp: datetime.datetime = Field(
        description="Timestamp of when the measurement was taken"
    )
    type: DataValueType = Field(description="The type of the data value taken")
    position: Optional[str] = Field(  # noqa: UP007
        description="Latitude and longitude GPS coordinates of the sensor"
    )
    value: int = Field(description="The actual value of the measurement taken")
    year: Optional[int] = Field(  # noqa: UP007
        default=None,
        description="Year when the measurement was taken"
        ", not included with newer data, prefer stamp",
    )


class OverallValues(BaseModel):
    no2: int
    o3: int
    pm25: int
    pm10: int
    temperature: int
    humidity: int
    pressure: int
    noise_dba: int


class Overall(BaseModel):
    city_name: str = Field(alias="cityName", description="The city name")
    values: OverallValues = Field(description="The overall values for the city")



from .client import PulseEcoClient
from .enums import AveragePeriod, DataValueType, SensorStatus, SensorType

__all__ = [
    "PulseEcoClient",
    "AveragePeriod",
    "DataValueType",
    "SensorStatus",
    "SensorType",
]
